import typing

from tottle.api.abc import ABCAPI
from tottle.api.request_rescheduler import ABCRequestRescheduler, BlockingRequestRescheduler
from tottle.api.request_validator import DEFAULT_REQUEST_VALIDATORS, ABCRequestValidator
from tottle.api.response_validator import DEFAULT_RESPONSE_VALIDATORS, ABCResponseValidator
from tottle.http import ABCSessionManager, SingleSessionManager, AiohttpClient
from tottle.modules import logger
from tottle.types.categories import APICategories

Token = typing.Union[str, typing.NewType("Token", str)]
APIRequest = typing.NamedTuple("APIRequest", [("method", str), ("data", dict)])


class API(ABCAPI, APICategories):
    API_URL = "https://api.telegram.org/"
    APIRequest = APIRequest

    def __init__(
        self,
        token: Token,
        ignore_errors: bool = False,
        session_manager: typing.Optional[SingleSessionManager] = None,
        request_rescheduler: typing.Optional[ABCRequestRescheduler] = None,
    ):
        self.token = token
        self.ignore_errors = ignore_errors
        self.request_url = self.API_URL + f"bot{self.token}/"
        self.http: ABCSessionManager = session_manager or SingleSessionManager(AiohttpClient)
        self.request_rescheduler = request_rescheduler or BlockingRequestRescheduler()
        self.response_validators: typing.List[ABCResponseValidator] = DEFAULT_RESPONSE_VALIDATORS
        self.request_validators: typing.List[ABCRequestValidator] = DEFAULT_REQUEST_VALIDATORS

    async def request(self, method: str, data: typing.Optional[dict] = None) -> dict:
        """Makes a single request opening a session"""
        data = await self.validate_request(data)

        async with self.http as session:
            response = await session.request_text(
                "POST",
                url=self.request_url + method,
                json=data,
            )

            logger.debug("Request {} with {} data returned {}".format(method, data, response))
            return await self.validate_response(method, data, response)

    async def request_many(
            self, requests: typing.Iterable[APIRequest]  # type: ignore
    ) -> typing.AsyncIterator[dict]:
        """ Makes many requests opening one session """
        async with self.http as session:
            for request in requests:
                method, data = request.method, await self.validate_request(request.data)  # type: ignore
                response = await session.request_text(
                    "POST",
                    url=self.request_url + method,
                    json=data,
                )
                logger.debug("Request {} with {} data returned {}".format(method, data, response))
                yield await self.validate_response(method, data, response)

    async def validate_response(
            self, method: str, data: dict, response: typing.Union[dict, str]
    ) -> typing.Union[typing.Any, typing.NoReturn]:
        """ Validates response from Telegram,
        to change validations change API.response_validators (list of ResponseValidator's) """
        for validator in self.response_validators:
            response = await validator.validate(method, data, response, self)
        logger.debug("API response was validated")
        return response  # type: ignore

    async def validate_request(self, request: dict) -> dict:
        """ Validates requests from Telegram,
        to change validations change API.request_validators (list of RequestValidator's) """
        for validator in self.request_validators:
            request = await validator.validate(request)
        logger.debug("API request was validated")
        return request  # type: ignore

    @property
    def api_instance(self) -> "API":
        return self
