from abc import ABC
from typing import List, Callable, Any, Optional

from tottle.api.abc import ABCAPI
from tottle.dispatch.dispenser.abc import ABCStateDispenser
from tottle.dispatch.middlewares.abc import MiddlewareResponse
from tottle.dispatch.handlers import ABCHandler
from tottle.dispatch.views import ABCDispenseView
from tottle.dispatch.middlewares import BaseMiddleware
from tottle.dispatch.return_manager import BotMessageReturnHandler
from tottle.modules import logger
from tottle.tools.dev_tools.mini_types.bot.message import message_min
from tottle.types.enums import BotUpdateType, get_update_type

DEFAULT_STATE_KEY = "id"


class ABCMessageView(ABCDispenseView, ABC):
    def __init__(self):
        self.state_source_key = DEFAULT_STATE_KEY
        self.handlers: List["ABCHandler"] = []
        self.middlewares: List["BaseMiddleware"] = []
        self.default_text_approximators: List[Callable[[message_min], str]] = []
        self.handler_return_manager = BotMessageReturnHandler()

    async def process_update(self, update: dict) -> bool:
        update_type = get_update_type(update)
        return BotUpdateType(update_type) == BotUpdateType.MESSAGE

    async def handle_update(
        self, update: dict, ctx_api: "ABCAPI", state_dispenser: "ABCStateDispenser"
    ) -> Any:

        logger.debug("Handling update ({}) with message view".format(update.get("update_id")))
        context_variables = {}
        message = message_min(update, ctx_api)
        message.state_peer = await state_dispenser.cast(self.get_state_key(update))

        for text_ax in self.default_text_approximators:
            message.text = text_ax(message)

        for middleware in self.middlewares:
            response = await middleware.pre(message)
            if response == MiddlewareResponse(False):
                return
            elif isinstance(response, dict):
                context_variables.update(response)

        handle_responses = []
        handlers = []

        for handler in self.handlers:
            result = await handler.filter(message)
            logger.debug("Handler {} returned {}".format(handler, result))

            if result is False:
                continue

            elif isinstance(result, dict):
                context_variables.update(result)

            handler_response = await handler.handle(message, **context_variables)
            handle_responses.append(handler_response)
            handlers.append(handler)

            return_handler = self.handler_return_manager.get_handler(handler_response)
            if return_handler is not None:
                await return_handler(
                    self.handler_return_manager, handler_response, message, context_variables
                )

            if handler.blocking:
                break

        for middleware in self.middlewares:
            await middleware.post(message, self, handle_responses, handlers)


class MessageView(ABCMessageView):
    def get_state_key(self, update: dict) -> Optional[int]:
        return update.get("message").get("chat").get(self.state_source_key)
