from typing import Any, Dict, List, NamedTuple, Callable

from tottle.api.abc import ABCAPI
from tottle.dispatch.dispenser.abc import ABCStateDispenser
from tottle.dispatch.handlers import ABCHandler
from tottle.dispatch.middlewares import BaseMiddleware, MiddlewareResponse
from tottle.dispatch.return_manager.bot import BotMessageReturnHandler
from tottle.dispatch.views.abc import ABCView
from tottle.types.enums import BotUpdateType, get_update_type
from tottle.modules import logger


HandlerBasement = NamedTuple(
    "HandlerBasement", [("dataclass", Callable), ("handler", ABCHandler)]
)


class RawUpdateView(ABCView):
    def __init__(self):
        self.handlers: Dict[BotUpdateType, HandlerBasement] = {}
        self.middlewares: List["BaseMiddleware"] = []
        self.handler_return_manager = BotMessageReturnHandler()

    async def process_update(self, update: dict) -> bool:
        if BotUpdateType(get_update_type(update)) in self.handlers:
            return True

    async def handle_update(
        self, update: dict, ctx_api: "ABCAPI", state_dispenser: "ABCStateDispenser"
    ) -> Any:
        logger.debug("Handling update ({}) with raw view".format(update.get("update_id")))

        update_type = get_update_type(update)
        handler_basement = self.handlers[BotUpdateType(update_type)]
        update_model = handler_basement.dataclass(**update.get(update_type))

        if isinstance(update_model, dict):
            update_model["ctx_api"] = ctx_api
        else:
            setattr(update_model, "unprepared_ctx_api", ctx_api)

        context_variables = {}

        for middleware in self.middlewares:
            response = await middleware.pre(update_model)
            if response == MiddlewareResponse(False):
                return
            elif isinstance(response, dict):
                context_variables.update(response)

        result = await handler_basement.handler.filter(update_model)
        logger.debug("Handler {} returned {}".format(handler_basement.handler, result))

        if result is False:
            return

        elif isinstance(result, dict):
            context_variables.update(result)

        handler_response = await handler_basement.handler.handle(update_model, **context_variables)

        return_handler = self.handler_return_manager.get_handler(handler_response)
        if return_handler is not None:
            await return_handler(
                self.handler_return_manager, handler_response, update_model, context_variables,
            )

        for middleware in self.middlewares:
            await middleware.post(
                update_model, self, [handler_response], [handler_basement.handler]
            )
