from asyncio import AbstractEventLoop, get_event_loop
from typing import Iterable, NoReturn, Optional, Union

from tottle.api import API, ABCAPI
from tottle.dispatch.dispenser.default import DefaultStateDispenser
from tottle.framework.bot.labeler import BotLabeler, ABCBotLabeler
from tottle.framework.abc import ABCFramework
from tottle.polling import BotPolling, ABCPolling
from tottle.dispatch import BotRouter, ABCRouter
from tottle.tools.dev_tools.loop_wrapper.loop_wrapper import LoopWrapper
from tottle.modules import logger
from tottle.exception_factory import ABCErrorHandler, ErrorHandler


class Bot(ABCFramework):
    def __init__(
        self,
        token: Optional[str] = None,
        api: Optional[ABCAPI] = None,
        polling: Optional[ABCPolling] = None,
        router: Optional["ABCRouter"] = None,
        labeler: Optional["ABCBotLabeler"] = None,
        loop: Optional[AbstractEventLoop] = None,
        loop_wrapper: Optional[LoopWrapper] = None,
        error_handler: Optional["ABCErrorHandler"] = None,
        task_each_update: bool = True,
    ):
        self.api: Union[ABCAPI, API] = API(token) if token else api  # type: ignore
        self.apis: Optional[Iterable["ABCAPI"]] = None
        self.error_handler = error_handler or ErrorHandler()
        self.loop_wrapper = loop_wrapper or LoopWrapper()
        self.labeler = labeler or BotLabeler()
        self.state_dispenser = DefaultStateDispenser()
        self._polling: "BotPolling" = polling or BotPolling(self.api)
        self._router: "BotRouter" = router or BotRouter()
        self._loop = loop
        self.task_each_update = task_each_update

    async def run_polling(
        self, custom_polling: Optional[ABCPolling] = None, apis_count: int = 0
    ) -> NoReturn:
        polling = custom_polling or self.polling
        logger.info("Polling will be started!")

        async for update in polling.listen():  # type: ignore
            logger.debug("New update: {}", update)

            if self.apis:
                self.api = self.apis[apis_count]

            if not self.task_each_update:
                await self.router.route(update, self.api)
            else:
                self.loop.create_task(self.router.route(update, self.api))

    def run_forever(self):
        logger.info("Loop will be ran forever")
        self.loop_wrapper.add_task(self.run_polling())
        self.loop_wrapper.run_forever(self.loop)

    @property
    def on(self) -> "ABCBotLabeler":
        return self.labeler

    @property
    def polling(self) -> "ABCPolling":
        return self._polling.construct(self.api)

    @property
    def router(self) -> "ABCRouter":
        return self._router.construct(
            views=self.labeler.views(),
            state_dispenser=self.state_dispenser,
            error_handler=self.error_handler,
        )

    @property
    def loop(self) -> AbstractEventLoop:
        if self._loop is None:
            self._loop = get_event_loop()
        return self._loop

    @loop.setter
    def loop(self, new_loop: AbstractEventLoop):
        self._loop = new_loop

    @router.setter
    def router(self, new_router: "ABCRouter"):
        self._router = new_router

    @polling.setter
    def polling(self, value):
        self._polling = value
